from casadi import *
from casadi.tools import *
from plotter import *
from pylab import *

"""
       NOTE: if you use spyder,
           make sure you open a Python interpreter
                 instead of an IPython interpreter
           otherwise you wont see any plots
"""


N = 20      # Control discretization
T = 10.0    # End time
h = T/N
M = 10      # Number of RK4 steps

# Declare variables (use scalar graph)
u  = SX.sym("u")    # control
x  = SX.sym("x",2)  # states

# Formulate the ODE
xdot = vertcat(((1-x[1]**2)*x[0]-x[1]+u, x[0]))
f = SXFunction([x,u],[xdot])
f.init()

# Discrete time dynamics
F = simpleRK(f, M)
F.init()

# System dynamics
xdot = vertcat( [(1 - x[1]**2)*x[0] - x[1] + u, x[0]] )
f = SXFunction([x,u],[xdot])
f.setOption("name","f")
f.init()

# Define NLP variables
W = struct_symMX([
      (
        entry("X",shape=(2,1),repeat=N+1),
        entry("U",shape=(1,1),repeat=N)
      )
])
w = MX(W)

# NLP constraints
g = []

# Build up a graph of integrator calls
for k in range(N):
    # Call the integrator
    [x_next_k] = F([ W["X",k], W["U",k], h ])

    # Append continuity constraints
    g.append(x_next_k - W["X",k+1])

# Concatenate constraints
g = vertcat(g)

# Objective function
obj = mul(w.T, w)

# Create an NLP solver object
nlp = MXFunction(nlpIn(x=W),nlpOut(f=obj,g=g))
nlp_solver = NlpSolver("ipopt", nlp)
nlp_solver.setOption("linear_solver", "mumps")
nlp_solver.init()

# All constraints are equality constraints in this case
nlp_solver.setInput(0, "lbg")
nlp_solver.setInput(0, "ubg")

# Construct and populate the vectors with
# upper and lower simple bounds
w_min = W(-inf)
w_max = W( inf)

# Control bounds
w_min["U",:] = -1
w_max["U",:] = 1

w_k = W(0)
ts = linspace(0,T,N+1)
plotter = Plotter(ts)
t = 0
x_current = array([1,0])
while True:
  w_min["X",0] = x_current
  w_max["X",0] = x_current

  # Pass data to NLP solver
  nlp_solver.setInput(w_k,"x0")
  nlp_solver.setInput(w_min,"lbx")
  nlp_solver.setInput(w_max,"ubx")
   
  # Solve the OCP
  nlp_solver.evaluate()
    
  # Extract from the solution the first control
  sol = W(nlp_solver.getOutput("x"))
  u_nmpc = sol["U",0]

  # Plot the solution
  plotter.show(t,x_current,sol)
  import sys
  sys.stdout.write('Waiting for your input (<enter>, "quit|clip|clear", or numbers ):')
  wait = raw_input()
  if "quit" in wait:
    break
  if "clear" in wait:
    plotter.clear()
  if "clip" in wait:
    plotter.toggleClipping()
  try:  # Easier to Ask Forgiveness than Permission
    x_current[:] = array(map(float,wait.split(" ")))
  except:
    pass
    
  # Simulate the system with this control
  [x_current] = F([x_current, u_nmpc, h])
  
  t += T/N
  # Shift the time to have a better initial guess
  # For the next time horizon
  w_k["X",:-1] = sol["X",1:]
  w_k["U",:-1] = sol["U",1:]
  w_k["X",-1] = sol["X",-1]
  w_k["U",-1] = sol["U",-1]
  
